#! /usr/bin/perl

# Antoni Oliver 2004
# aoliver@lsi.upc.es
# Aquest programa es pot fer servir, modificar i distribuir lliurement sempre que es faci referència a l'autor
# Modificado por Andrés Chandía 2008
# andres@chandia.net
# Este programa se puede utilizar, modificar i distribuir libremente siempre que se haga referencia a su autor (Antoni Oliver).

use strict;
use Getopt::Long;

#valores por defecto
my $sa;
my $st;
my $rec;
my $pa;
my $minraices=10;
my $minterm=5;
my $maxterm=100;
my $ntp=1;
my $maxcar=5;
my $mincar=1;
my $entrada;
my $salida;
my $help;

GetOptions( "sa!" => \$sa,
	    "st!" => \$st,
	    "mia:i" => \$minraices,
	    "mit:i" => \$minterm,
	    "mat:i" => \$maxterm,
	    "ntp:i" => \$ntp,
	    "mac:i" => \$maxcar,
	    "mic:i" => \$mincar,
	    "e=s" => \$entrada,
            "s=s"  => \$salida,
	    "pa:s" => \$pa,
	    "rec!" => \$rec,
	    "help" => \$help);

if ($help)
{
    print "Ayuda;\n";
    print "Programa par descubrir la morfología. Los parámetros son:\n";
    print "mia (integer): número mínimo de raíces en un paradigma. Opcional, valor por defecto 10.\n";
    print "mit (integer): número mínimo de terminaciones en un paradigma. Opcional: valor por defecto: 5\n";
    print "mat (integer): número máximo de terminaciones en un paradigma. Opcional: valor por defecto: 100\n";
    print "ntp (integer): número de mejores terminaciones por cada iteración. Opcional: valor por defecto: 1\n";
    print "mac (integer): número máximo de caracteres para una terminación. Opcional: valor por defecto: 5\n";
    print "mic (integer): número mínimo de caracteres para una terminación. Opcional: valor por defecto: 1\n";
    print "e (string): nombre del fichero de entrada. Obligatorio.\n";
    print "s (string): nombre del fichero de salida. Obligatorio.\n";
    print "pa (string): nombre del fichero de los paradigmas acceptados. Opcional.\n";
    print "--sa: activa la simplificación por raíces\n";
    print "--st: activa la simplificación por terminaciones\n";  
    print "--rec: activa la recuperación de raíces\n";
    print "help: muestra esta ayuda\n";
    print "Final de la ayuda";
    exit;
}
unless ($entrada) {print "Se debe especificar un fichero de entrada\n";exit;}
unless ($salida) {print "Se debe especificar un fichero de salida\n";exit;}

open(IN,$entrada)||die "No puedo abrir el fichero de entrada: $!\n";
open(OUT,">$salida")||die "No puedo abrir el fichero de salida: $!\n";



#proceso
my %palabra=();
my $totalpalabras=0;
my %morfologia=();
my %temp_morfologia=();
my @palabras=();
$totalpalabras=cargarpalabras();

#si hay paradigmes aceptados, eliminamos las palabras ya descritas del hash palabra

if ($pa)
{
    open (PA,$pa);
    while (<PA>)
    {
	chomp;
	(my $ts,my $as)=split(/\t/);
	my @ts=split(/:/,$ts);
	my @as=split(/:/,$as);
	foreach my $a (@as)
	{
	    foreach my $t (@ts)
	    {
		if ($t=="0"){$t=""}
		my $f=$a.$t;
		delete($palabra{$f});
	    }
	}
    }
}
    
	



print "TOTAL palabras: $totalpalabras\n";

for (my $numcar=$maxcar;$numcar>=$mincar;$numcar--)
{
    %temp_morfologia=();
    @palabras=keys(%palabra);
    print "NUMCAR $numcar\n";
    my @mpt=mpt($ntp,$numcar);
    foreach my $mpt (@mpt)
    {
	%temp_morfologia=(%temp_morfologia,division($mpt));
    }


    %temp_morfologia=simplificarnum(%temp_morfologia);
    if ($sa){%temp_morfologia=simplificarraices(%temp_morfologia);}
    if ($st){%temp_morfologia=simplificarterminaciones(%temp_morfologia);}
    #en cada pasada conservamos la más productiva

    my @bestterms=(sort{productivitad($b,$temp_morfologia{$b})<=>productividad($a,$temp_morfologia{$a})}(keys(%temp_morfologia)));
    my $bestterm=$bestterms[0];

    if ($bestterm eq undef)
    {
	print "No hay\n";next;
    }
    else
    {
	if ($rec)
	{
	    my $raicesrec=recupera($bestterm);
	    $temp_morfologia{$bestterm}=$raicesrec;
	}
        print OUT "$bestterm\t$temp_morfologia{$bestterm}\n";
	borrar($bestterm,$temp_morfologia{$bestterm});
	redo;
    }
		      
			  


}

sub cargarpalabras
{
    my $totalpalabras=0;
    while (<IN>)
    {
	chomp;
	++$palabra{$_};
	++$totalpalabras;
    }
    return($totalpalabras);
}

sub mpt
{
    my $ntp=$_[0];
    my $numcar=$_[1];
    #calculamos la terminación más probable de una cierta cantidad de letras
    my %term=();
    foreach my $palabra (keys(%palabra))
    {
	my $lpar=length($palabra);
	if ($lpar > $numcar)
	{
	    my $term=substr($palabra,$lpar-$numcar,$numcar);
	    ++$term{$term};
	}
    }
    my $control=0;
    my @mpt=();
    foreach my $key (sort{$term{$b}<=>$term{$a}} (keys(%term)))
    {
	++$control;
	if ($control > $ntp) {last;}
	else{push(@mpt,$key);}
    }
    return(@mpt);
}

sub division
{
    my $terminacion=$_[0];
    my $terminaciones;
    my %temp_morfologia=();

    foreach my $palabra (keys (%palabra))
    {
	if ($palabra =~/$terminacion$/)
	{
	    my @tmp_words=();
	    my @tmp_term=();
	    push(@tmp_term,$terminacion);
	    my $palabra2=$palabra;
	    $palabra2=~s/$terminacion$//;
	    
	    @tmp_words =grep /^$palabra2/,@palabras;
	    unshift(@tmp_words,$palabra);
	    foreach my $tmp_word (@tmp_words)
	    {
		if ($tmp_word ne $palabra)
		{
		    if ($tmp_word eq $palabra2) 
		    {
			$tmp_word="0";
		    }
		    else
		    {
			$tmp_word=~s/^$palabra2//;
		    }
		    push(@tmp_term,$tmp_word);
		}
		@tmp_term=sort(@tmp_term);
		$terminaciones=join(":",@tmp_term);
		$temp_morfologia{$terminaciones}.=":".$palabra2;
		
	    }
	}
    }
    foreach my $key (keys(%temp_morfologia))
    {
	$temp_morfologia{$key}=~s/^://;
	#verificamos que no haya raíces repetidas
	my %controlraices=();
	my @controlraices=split(/:/,$temp_morfologia{$key});
	foreach my $controlraiz (sort(@controlraices))
	{
	    ++$controlraices{$controlraiz};
	}
	@controlraices=sort(keys(%controlraices));
	my $controlraiz=join(":",@controlraices);
	#verificamos que no haya terminaciones repetidas
	my %controlterms=();
	my @controlterms=split(/:/,$key);
	foreach my $controlterm (sort(@controlterms))
	{
	    ++$controlterms{$controlterm};
	}
	@controlterms=sort(keys(%controlterms));
	my $controlterm=join(":",@controlterms);
	delete($temp_morfologia{$key});
	$temp_morfologia{$controlterm}=$controlraiz;


    }
    return(%temp_morfologia);

}

sub simplificarnum
{
    my %temp_morfologia=@_;
    foreach my $key (keys(%temp_morfologia))
    {
	my @t=split(/:/,$key);
	my $nt=@t;
	my @a=split(/:/,$temp_morfologia{$key});
	my $na=@a;
	if (($na<$minraices)||($nt<$minterm)||($nt>$maxterm))
	{
	    delete($temp_morfologia{$key});
	}
    }


    return(%temp_morfologia);
}

sub simplificarraices
{
    #no pueden acabar todas las raices con la misma letra
    my %temp_morfologia=@_;
    foreach my $key (keys(%temp_morfologia))
    {
	my @raices=split(/:/,$temp_morfologia{$key});
	my $iguala=1;
	my $raiz=$raices[0];
	my $ula1=substr($raiz,length($raiz)-1,1);

	foreach $raiz (@raices)
	{
	    my $ula=substr($raiz,length($raiz)-1,1);
	    if ($ula ne $ula1) {$iguala=0;last}
	}
	if ($iguala == 1)
	{

	    delete($temp_morfologia{$key});
	}
    }
    return(%temp_morfologia);
}

sub simplificarterminaciones
{
    #las terminaciones no pueden comenzar por la misma letra
    my %temp_morfologia=@_;
    foreach my $key (keys(%temp_morfologia))
    {

	my @terminaciones=split(/:/,$key);
	my $igualt=1;
	my $term=$terminaciones[0];
	my $plt1=substr($term,0,1);
	foreach $term (@terminaciones)
	{
	    my $plt=substr($term,0,1);
	    if ($plt ne $plt1) {$igualt=0;last};
	}
	if ($igualt == 1)
	{
	    delete($temp_morfologia{$key});
	}
    }
    return(%temp_morfologia);
}

sub productividad
{
    my $elementos1=$_[0];
    my $elementos2=$_[0];
    my @elementos1=split(/:/,$elementos1);
    my @elementos2=split(/:/,$elementos2);
    my $ne1=@elementos1;
    my $ne2=@elementos2;
    my $productividad=$ne1*$ne2;
    return($productividad);
}

sub borrar
{
    my $terminaciones=$_[0];
    my $raices=$_[1];
    my @terminaciones=split(/:/,$terminaciones);
    my @raices=split(/:/,$raices);
    foreach my $a (@raices)
    {
	foreach my $t (@terminaciones)
	{
	    if ($t eq "0"){$t=""}
	    my $forma=$a.$t;
	    delete($palabra{$forma});
	}
    }
}

sub recupera
{
    my $terms=$_[0];
    my @terms=split(/:/,$terms);
    my @formas=();
    my %raices=();
    if ($terms[0] eq "0"){$terms[0]=""}
    my @raices=();
    foreach my $palabra (keys(%palabra))
    {
	if ($palabra=~/$terms[0]$/)
	{
	    my $raiz=$palabra;
	    $raiz=~s/$terms[0]$//;
	    my $incloure=1;
	    foreach my $t (@terms)
	    {
		my $forma=$raiz.$t;
		if ($palabra{$forma} eq undef)
		{
		    $incloure=0;
		    last;
		}
	    }
	    if ($incloure == 1) {++$raices{$raiz}}
	}
    }
    @raices=sort(keys(%raices));
    my $raices=join(":",@raices);
    return($raices);
}
